"""Base exceptions module"""


class RPCError(Exception):
    """Base error class for RPC protocols"""

    message: str = None

    def __init__(self, *args):
        super().__init__(*args)
        if len(args) > 0:
            self.message = str(args[0])

    def as_dict(self) -> dict[str, str]:
        return {
            "error": self.__class__.__name__,
            "message": str(self.message or self),
        }


class SerializationError(RPCError):
    """Serialization error"""


class DeserializationError(RPCError):
    """Deserialization error"""


class BadRequestError(RPCError):
    """Bad request error"""


class NotFoundError(RPCError):
    """Not found error"""


class InvalidParametersError(RPCError):
    """Invalid parameters error"""


class AuthenticationFailedError(RPCError):
    """Authentication failed"""


class InvalidTokenError(AuthenticationFailedError):
    """Invalid token error"""


class ForbiddenError(AuthenticationFailedError):
    """Forbidden error"""


class InvalidPermissionsError(ForbiddenError):
    """Invalid permissions error"""


class InternalServerError(RPCError):
    """Server error"""



from __future__ import annotations

import re
from typing import Iterable
from typing import Optional

from drakaina import ENV_AUTH_PAYLOAD
from drakaina import ENV_AUTH_SCOPES
from drakaina import ENV_IS_AUTHENTICATED
from drakaina import ENV_USER
from drakaina import ENV_USER_ID
from drakaina._types import ASGIReceive
from drakaina._types import ASGIScope
from drakaina._types import ASGISend
from drakaina._types import WSGIApplication
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse
from drakaina.contrib.jwt import SUPPORTED_ALGORITHMS
from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.types import RevokeChecker
from drakaina.contrib.jwt.types import ScopesGetter
from drakaina.contrib.jwt.types import TokenGetter
from drakaina.contrib.jwt.types import UserGetter
from drakaina.contrib.jwt.utils import decode_jwt_token
from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import ForbiddenError
from drakaina.middleware.base import BaseMiddleware
from drakaina.utils import get_cookies
from drakaina.utils import iterable_str_arg


class JWTAuthenticationMiddleware(BaseMiddleware):
    """The middleware supporting JWT tokens.

    :param app: WSGI Application.
    :type app: WSGIApplication
    :param algorithms: List of supported algorithms.
    :type algorithms: Iterable[str]
    :param secret_key: A secret key for signature verification.
    :type secret_key: str
    :param public_key: Public key for signature verification in
        asymmetric algorithms.
    :type public_key: str
    :param credentials_required: If `True`, the request is expected to
        contain credential data. If they are missing, an exception is raised.
        Default: `True`.
    :type credentials_required: bool
    :param prefix: The token prefix in the authorization headers.
        Default: `"Bearer"`.
    :type prefix: str
    :param use_cookies: If true, the token will be searched in the cookie
        headers for the specified key. Default: `False`.
    :type use_cookies: bool
    :param cookie_key: The key by which the token is stored in the cookie
        headers.
    :type cookie_key: str
    :param decode_options: The dictionary is passed as is as the `options`
        argument to the `pyjwt.decode` function. Default: `{
            "require": [],  # Require claims
            "verify_iss": False,  # Issuer
            "verify_aud": False,  # Audience
            "verify_exp": True,   # Expiry
            "verify_nbf": False,  # Not Before
            "verify_iat": True,   # Issued at
            "verify_jti": False,  # JWT ID
        }`
    :type decode_options: Optional[dict[str, bool | Iterable[str]]]
    :param verify_values: A dictionary of key-value pairs, where key is
        the name of the key in the token payload, and the value to check can be
        a specific value, a list of possible values, or a re.Pattern object
        (the result of re.compile).
    :type verify_values: Optional[dict[str, str | Iterable[str] | re.Pattern]]
    :param leeway_value: The value of leeway in seconds. Default: `0`.
    :type leeway_value: int | float
    :param user_id_field: The name of the key in the token payload to store
        the user ID. Default: `"user_id"`.
    :type user_id_field: str
    :param token_getter: Callable for an alternative way to get a token.
    :type token_getter: TokenGetter
    :param user_getter: Callable to retrieve a user object.
    :type user_getter: UserGetter
    :param scopes_getter: Callable to get permission scopes.
    :type scopes_getter: ScopesGetter
    :param revoke_checker: Callable to verify that the token is revoked.
    :type revoke_checker: RevokeChecker
    :param kwargs: Other arguments to pass to the constructor of the base class.

    """

    __slots__ = (
        "_algorithms",
        "__secret_key",
        "__public_key",
        "_credentials_required",
        "_prefix",
        "_cookie_key",
        "_decode_options",
        "_verify_values",
        "_leeway",
        "_user_id_field",
        "get_token",
        "get_user",
        "is_revoked",
        "get_scopes",
    )

    def __init__(
        self,
        app: WSGIApplication,
        algorithms: str | Iterable[str] = SUPPORTED_ALGORITHMS,
        secret_key: str = None,
        public_key: str = None,
        credentials_required: bool = True,
        prefix: str = "Bearer",
        use_cookies: bool = False,
        cookie_key: str = "jwt",
        decode_options: dict[bool | list[str]] = None,
        verify_values: dict[str, str | Iterable[str] | re.Pattern] = None,
        leeway_value: int | float = 0,
        user_id_field: str = "user_id",
        token_getter: Optional[TokenGetter] = None,
        user_getter: Optional[UserGetter] = None,
        scopes_getter: Optional[ScopesGetter] = None,
        revoke_checker: Optional[RevokeChecker] = None,
        **kwargs,
    ):
        super().__init__(app, **kwargs)

        self._algorithms = iterable_str_arg(algorithms)
        self.__secret_key = secret_key
        self.__public_key = public_key
        self._credentials_required = credentials_required
        self._prefix = prefix
        self._cookie_key = cookie_key
        self._decode_options = (
            decode_options
            if decode_options is not None
            else {
                "require": [],
                "verify_iss": False,
                "verify_aud": False,
                "verify_exp": True,
                "verify_nbf": False,
                "verify_iat": True,
                "verify_jti": False,
            }
        )
        self._verify_values = verify_values if verify_values is not None else {}
        self._leeway = leeway_value
        self._user_id_field = user_id_field

        if callable(token_getter):
            self.get_token = token_getter
        elif use_cookies:
            self.get_token = self._token_from_cookies
        else:
            self.get_token = self._token_from_auth_header

        self.is_revoked = revoke_checker
        self.get_user = user_getter
        self.get_scopes = scopes_getter

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        if environ["REQUEST_METHOD"] == "OPTIONS":
            return self.app(environ, start_response)

        token = self.get_token(environ)

        if not token and self._credentials_required:
            raise AuthenticationFailedError("Credential required")

        if token is not None:
            token_payload = decode_jwt_token(
                token=token,
                algorithms=self._algorithms,
                signing_key=self.__secret_key,
                verify_key=self.__public_key,
                verify=True,
                decode_options=self._decode_options,
                verify_values=self._verify_values,
                leeway=self._leeway,
            )

            if callable(self.is_revoked):
                if self.is_revoked(environ, token_payload):
                    raise ForbiddenError("Token is revoked")

            environ[ENV_USER_ID] = token_payload[self._user_id_field]
            environ[ENV_IS_AUTHENTICATED] = True
            environ[ENV_AUTH_PAYLOAD] = token_payload

            if callable(self.get_user):
                environ[ENV_USER] = self.get_user(environ, token_payload)
            if callable(self.get_scopes):
                environ[ENV_AUTH_SCOPES] = self.get_scopes(
                    environ,
                    token_payload,
                )

        return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        await self.app(scope, receive, send)

    def _token_from_auth_header(
        self,
        request: ASGIScope | WSGIEnvironment,
    ) -> Optional[str]:
        auth_header = request.get("HTTP_AUTHORIZATION")

        if auth_header is None:
            return None

        try:
            parts = auth_header.strip().split(" ")
        except ValueError:
            raise AuthenticationFailedError("Invalid `Authorization` header")

        if len(parts) == 0 or parts[0] not in self._prefix:
            return None

        if len(parts) != 2:
            raise InvalidJWTTokenError(
                "The `Authorization` header must contain two values "
                "separated by a space",
            )

        return parts[1]

    def _token_from_cookies(
        self,
        request: ASGIScope | WSGIEnvironment,
    ) -> Optional[str]:
        cookie_header = request.get("HTTP_COOKIE")

        if cookie_header is None:
            return None

        cookies = get_cookies(cookie_header)

        return cookies.get(self._cookie_key)



class ASGIHandler:
    """"""

    def __init__(self, route: str):
        self.route = route

        # if prefix:
        #     assert prefix.startswith("/"), "A path prefix must start with '/'"
        #     assert not prefix.endswith(
        #         "/"
        #     ), "A path prefix must not end with '/',
        #     as the routes will start with '/'"

    """http scope:
    {
        'type': 'http.request',
        'scheme': 'http',
        'root_path': '',
        'server': ('127.0.0.1', 8000),
        'http_version': '1.1',
        'method': 'GET',
        'path': '/',
        'headers': [
            [b'host', b'127.0.0.1:8000'],
            [b'user-agent', b'curl/7.51.0'],
            [b'accept', b'*/*']
        ]
    }
    """

    async def __call__(self, scope: dict, receive, send):
        import sys

        if scope["type"] == "http":
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain"]],
                }
            )
            version = f"{sys.version_info.major}.{sys.version_info.minor}"
            message = (
                f"Hello world! From Uvicorn with Gunicorn. "
                f"Using Python {version}".encode("utf-8")
            )
            await send({"type": "http.response.body", "body": message})
        elif scope["type"] == "websocket":
            await send({"type": "websocket.send"})



"""
Some recommendations for implementation
https://www.simple-is-better.org/json-rpc/transport_sockets.html

====
This module holds all of the functions and classes necessary for
initiating a JSONRPC-TCP Server.
"""
from __future__ import print_function

import json
import socket
import sys
import threading
import time
import traceback

EncryptionMissing, ProtocolError = Exception, Exception


class Server(object):
    """
    This class is the basic Server object. It should be instantiated
    with a (host, port) tuple (and an optional handler), and then
    the Handler subclasses / functions should be attached through the
    add_handler method.
    """

    _shutdown = False

    def __init__(self, addr, handler=None, pool=10):
        # if config.secret and not config.crypt:
        #     raise EncryptionMissing("No encrpytion library found.")
        self.addr = addr
        self.socket = None
        self.threads = []
        # Pool not actually implemented yet
        self.pool = pool
        self.json_request = JSONRequest(self)
        # if handler:
        #     assert hasattr(handler, '__call__') or \
        #            issubclass(handler, Handler)
        #     self.json_request.add_handler(handler)

    def serve(self):
        """
        This starts the server -- it blocks, so if there are other
        tasks that need to be performed after the server is started,
        threading / multiprocessing will need to be employed.
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.addr)
        self.socket.listen(100)  # (config.max_queue)
        self.wait()

    def wait(self):
        """The principle wait cycle."""
        while True:
            if self._shutdown:
                break
            clientsock, addr = self.socket.accept()
            args = (clientsock, addr)
            target = self.json_request.process
            thread = threading.Thread(target=target, args=args)
            thread.daemon = True
            thread.start()
            self.threads.append(thread)
            self.check_threads()

        sys.stdout.write("Shutting down...")
        for thread in self.threads:
            thread.join()
        sys.stdout.write("done.\n")

    def shutdown(self):
        """
        Attempts to shutdown the server.
        TODO: Make this work quickly and properly.
        """
        self._shutdown = True
        self.socket.close()

    def check_threads(self):
        """
        Check the thread list for dead threads and finished
        threads.
        """
        for thread in self.threads:
            if not thread.isAlive():
                thread.join()
                self.threads.remove(thread)

    def add_handler(self, method, name=None):
        """Just a wrapper around JSONRequest.add_handler"""
        self.json_request.add_handler(method, name)


class JSONRequest(object):
    """
    This is the class that handles individual requests passed
    from the server wait cycle.
    """

    def __init__(self, server):
        self.server = server
        self.handlers = {}

    def add_handler(self, method, name=None):
        """
        Attach a handler to the request object. It must be either
        callable, or a subclass of Handler.
        """
        # if isclass(method):
        #     assert issubclass(method, Handler)
        #     # If it's an actual Handler subclass
        #     handler_instance = method(self)
        #     for hname, method in handler_instance._handlers.iteritems():
        #         if name:
        #             hname = "%s.%s" % (name, hname)
        #         self.handlers[hname] = method
        # else:
        if not name:
            name = method.__name__
        assert hasattr(method, "__call__")
        self.handlers[name] = method

    def get_handler(self, name):
        """Check for an attached handler and return it."""
        if name in self.handlers:
            return self.handlers[name]
        return None

    def process(self, sock, addr):
        """Just a wrapper for ProcessRequest."""
        request = ProcessRequest(self)
        request.process(sock, addr)


class ProcessRequest(object):
    """
    This is the class that handles an actual request, passing it through
    the Handler and parsing the response.
    """

    socket_error = False

    def __init__(self, json_request):
        self.json_request = json_request
        self.socket = None
        self.client_address = None

    def process(self, sock, addr):
        """
        Retrieves the data stream from the socket and validates it.
        """
        self.socket = sock
        self.socket.settimeout(30)  # (config.timeout)
        self.client_address = addr
        requestlines = []
        while True:
            data = self.get_data()
            if not data:
                break
            requestlines.append(data)
            if len(data) < 4096:  # config.buffer:
                break
        request = "".join(requestlines)
        response = ""
        crypt_error = False
        # if config.secret:
        #     crypt = config.crypt.new(config.secret)
        #     try:
        #         request = crypt.decrypt(request)
        #     except ValueError:
        #         crypt_error = True
        #         error = ProtocolError(-32700, "Could not decrypt request.")
        #         response = json.dumps(error.generate_error())
        # history.request = request
        # logger.debug("SERVER | REQUEST: %s" % request)
        if self.socket_error:
            self.socket.close()
        else:
            if not crypt_error:
                response = self.parse_request(request)
            # history.response = response
            # logger.debug("SERVER | RESPONSE: %s" % response)
            # if config.secret:
            #     length = config.crypt_chunk_size
            #     pad_len = length - (len(response) % length)
            #     response = crypt.encrypt("%s%s" % (response, " " * pad_len))
            self.socket.send(response)
        self.socket.close()

    def get_data(self):
        """Retrieves a data chunk from the socket."""
        try:
            data = self.socket.recv(4096)  # (config.buffer)
        except socket.timeout:
            # It may have finished sending without an error if
            # len(message) % buffer == 0.
            data = None
        except socket.error:
            self.socket_error = True
            data = None
        return data

    def parse_request(self, data):
        """Attempts to load the request, validates it, and calls it."""
        try:
            obj = json.loads(data)
        except ValueError:
            return json.dumps(ProtocolError(-32700).generate_error())
        if not obj:
            return json.dumps(ProtocolError(-32600).generate_error())
        batch = True
        if type(obj) is not list:
            batch = False
            obj = [
                obj,
            ]
        responses = []
        for req in obj:
            request_error = ProtocolError(-32600)
            if type(req) is not dict:
                responses.append(request_error.generate_error())
            elif "method" not in req.keys() or type(req["method"]) != str:
                responses.append(request_error.generate_error())
            else:
                result = self.parse_call(req)
                if req.has_key("id"):
                    response = generate_response(result, id=req.get("id"))
                    responses.append(response)
        if not responses:
            # It's either a batch of notifications or a single
            # notification, so return nothing.
            return ""
        else:
            if not batch:
                # Single request
                responses = responses[0]
            return json.dumps(responses)

    def parse_call(self, obj):
        """
        Parses a JSON request.
        """

        # Get ID, Notification if None
        # This is actually incorrect, as IDs can be null by spec (rare)
        # request_id = obj.get("id", None)

        # Check for required parameters
        jsonrpc = obj.get("jsonrpc", None)
        method = obj.get("method", None)
        if not jsonrpc or not method:
            return ProtocolError(-32600)

        # Validate parameters
        params = obj.get("params", [])
        if type(params) not in (list, dict):
            return ProtocolError(-32602)

        # Parse Request
        kwargs = {}
        if type(params) is dict:
            kwargs = params
            params = []
        handler = self.json_request.get_handler(method)
        # error_code = None
        message = None
        if handler:
            try:
                response = handler(*params, **kwargs)
                return response
            except Exception:
                # logger.error("Error calling handler %s" % method)
                message = traceback.format_exc().splitlines()[-1]
                error_code = -32603
        else:
            error_code = -32601
        return ProtocolError(error_code, message=message)


def generate_response(result, **kwargs):
    """
    TODO: Fix so that a request_id can be Null and not a Notification.
    """
    if type(result) is ProtocolError:
        return result.generate_error(**kwargs)
    else:
        response = {"jsonrpc": "2.0", "result": result}
        response.update(kwargs)
        return response


def start_server(host, port, handler):
    """
    Wrapper around Server that pre-threads it.
    """
    server = Server((host, port))
    server.add_handler(handler)
    server_thread = threading.Thread(target=server.serve)
    server_thread.daemon = True
    server_thread.start()
    return server


def test_server():
    """
    Creates a simple server to be tested against the test_client in
    the client module.
    """

    host, port = "", 8080

    def echo(message):
        """
        Test method, an example of how simple it *should* be.
        """
        return message

    def summation(*args):
        return sum(args)

    if "-v" in sys.argv:
        pass

        # config.verbose = True
        # logger.addHandler(logging.StreamHandler())
        # logger.setLevel(logging.DEBUG)

    server = Server((host, port))
    server.add_handler(echo)
    server.add_handler(echo, "tree.echo")
    server.add_handler(summation, "sum")
    server_thread = threading.Thread(target=server.serve)
    server_thread.daemon = True
    server_thread.start()

    print("Server running: %s:%s" % (host, port))

    try:
        while True:
            time.sleep(0.5)
    except KeyboardInterrupt:
        print("Finished.")
        sys.exit()


if __name__ == "__main__":
    from . import config

    config.verbose = True  # default is False
    config.timeout = 30  # default is 5
    config.buffer = 4096  # default is 1024
    # config.crypt = DES3  # default is AES if pycrypto is installed
    config.secret = "12345abcdef67890"

    test_server()



from __future__ import annotations

from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import MutableSequence
from collections.abc import Sequence
from sys import version_info
from typing import Any
from typing import Iterable
from typing import Type
from typing import Union

if version_info >= (3, 8):
    from typing import Literal
    from typing import Protocol
    from typing import TypedDict
else:
    from typing_extensions import Literal
    from typing_extensions import Protocol
    from typing_extensions import TypedDict

if version_info >= (3, 11):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired


"""
JSON-RPC types definitions
"""

JSONSimpleTypes = Union[str, int, float, bool, None]
JSONTypes = Union[
    JSONSimpleTypes,
    Mapping[str, JSONSimpleTypes],
    Sequence[JSONSimpleTypes],
]


class JSONRPCRequestObject(TypedDict):
    jsonrpc: Literal["2.0"]
    method: str
    params: NotRequired[list[JSONTypes] | dict[str, JSONTypes]]
    id: NotRequired[str | int | None]


class JSONErrorObject(TypedDict):
    code: int
    message: str
    data: NotRequired[JSONTypes]


class JSONRPCResponseObject(TypedDict):
    jsonrpc: Literal["2.0"]
    result: NotRequired[JSONTypes]
    error: NotRequired[JSONErrorObject]
    id: str | int | None


JSONRPCBatchRequestObject = Sequence[JSONRPCRequestObject]
JSONRPCRequest = Union[JSONRPCRequestObject, JSONRPCBatchRequestObject]
JSONRPCBatchResponseObject = Sequence[JSONRPCResponseObject]
JSONRPCResponse = Union[JSONRPCResponseObject, JSONRPCBatchResponseObject]


"""
WSGI types definitions
PEP 3333 – Python Web Server Gateway Interface
https://peps.python.org/pep-3333/
"""

WSGIEnvironmentKeys = Literal[
    # for CGI
    # https://datatracker.ietf.org/doc/html/draft-coar-cgi-v11-03
    "AUTH_TYPE",
    "CONTENT_LENGTH",
    "CONTENT_TYPE",
    "GATEWAY_INTERFACE",
    "PATH_INFO",
    "PATH_TRANSLATED",
    "QUERY_STRING",
    "REMOTE_ADDR",
    "REMOTE_HOST",
    "REMOTE_IDENT",
    "REMOTE_USER",
    "REQUEST_METHOD",
    "SCRIPT_NAME",
    "SERVER_NAME",
    "SERVER_PORT",
    "SERVER_PROTOCOL",
    "SERVER_SOFTWARE",
    # for WSGI
    "wsgi.errors",
    "wsgi.input",
    "wsgi.multiprocess",
    "wsgi.multithread",
    "wsgi.run_once",
    "wsgi.url_scheme",
    "wsgi.version",
]
# for framework needs
WSGIDrakainaKeys = Literal[
    "drakaina.app",
    "drakaina.is_authenticated",
]
WSGIEnvironment = MutableMapping[str, Any]
WSGIExceptionInfo = tuple[Type[BaseException], BaseException, Any]


class WSGIStartResponse(Protocol):
    def __call__(
        self,
        status: str,
        headers: MutableSequence[tuple[str, str]],
        exc_info: WSGIExceptionInfo | None = ...,
    ) -> Callable[[bytes], Any]:
        ...


WSGIResponse = Iterable[bytes]
WSGIApplication = Callable[[WSGIEnvironment, WSGIStartResponse], WSGIResponse]


class WSGIInputStream(Protocol):
    def read(self, size: int | None = None) -> bytes:
        ...

    def readline(self) -> bytes:
        ...

    def readlines(self, hint: Any | None) -> Iterable[bytes]:
        ...

    def __iter__(self) -> bytes:
        ...


class WSGIErrorsStream(Protocol):
    def flush(self) -> None:
        ...

    def write(self, s: str) -> None:
        ...

    def writelines(self, seq: Sequence[str]) -> None:
        ...


"""
ASGI types definitions
"""

ASGIScope = MutableMapping[str, Any]
ASGIMessage = MutableMapping[str, Any]

ASGIReceive = Callable[[], Awaitable[ASGIMessage]]
ASGISend = Callable[[ASGIMessage], Awaitable[None]]

ASGIApplication = Callable[[ASGIScope, ASGIReceive, ASGISend], Awaitable[None]]


"""
Helpful types
"""


class Comparator(Protocol):
    def __call__(
        self,
        required: Iterable[str],
        provided: str | Iterable[str],
    ) -> bool:
        ...


class ProxyRequest(MutableMapping):
    """A wrapper class for environment mapping.

    :param environment:

    """

    __slots__ = ("__environment",)

    def __init__(self, environment: ASGIScope | WSGIEnvironment):
        self.__environment = environment

    def __getitem__(self, item):
        return self.__environment[item]

    def __setitem__(self, key, value):
        self.__environment[key] = value

    def __delitem__(self, key):
        del self.__environment[key]

    def __iter__(self):
        return iter(self.__environment.keys())

    def __contains__(self, item):
        return item in self.__environment.keys()

    def __len__(self):
        return len(self.__environment)

    def keys(self):
        return self.__environment.keys()

    def values(self):
        return self.__environment.values()

    def items(self):
        return self.__environment.items()

    def get(self, key, default=None):
        return self.__environment.get(key, default)

    def clear(self):
        self.__environment.clear()

    def setdefault(self, key, default=None):
        self.__environment.setdefault(key, default)

    def pop(self, key, default=None):
        return self.__environment.pop(key, default)

    def popitem(self):
        return self.__environment.popitem()

    def copy(self):
        return self.__class__(self.__environment.copy())

    def update(self, *args, **kwargs):
        return self.__environment.update(*args, **kwargs)

    def __getattr__(self, item):
        if item in ["__environment", "_ProxyRequest__environment"]:
            super().__getattribute__(item)
        return self.__environment[item]

    def __setattr__(self, key, value):
        if key in ["__environment", "_ProxyRequest__environment"]:
            super().__setattr__(key, value)
        else:
            self.__environment[key] = value

    def __delattr__(self, item):
        del self.__environment[item]


AnyRequest = Union[ASGIScope, WSGIEnvironment, ProxyRequest]



from __future__ import annotations

import re
from datetime import datetime
from datetime import timedelta

try:
    from datetime import UTC
except ImportError:
    from datetime import tzinfo

    UTC = tzinfo(timedelta(minutes=0))  # noqa

from typing import Any
from typing import Iterable
from typing import Optional
from uuid import uuid4

from jwt import get_unverified_header
from jwt import InvalidAlgorithmError
from jwt import InvalidTokenError
from jwt import PyJWKClient
from jwt.api_jwt import decode_complete

from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.errors import JWTBackendError
from drakaina.contrib.jwt.errors import ValidationJWTTokenError
from drakaina.utils import iterable_str_arg


def datetime_utc_now() -> datetime:
    """Returns the current date/time in the UTC time zone.

    :return: The current date/time is in the UTC time zone.

    """
    return datetime.now(tz=UTC)


def datetime_to_timestamp(dt: datetime) -> int:
    """Converts the specified date/time to unixtime timestamp.

    :param dt: Date/time for conversion.
    :return: Unixtime timestamp.

    """
    return int(dt.timestamp())


def datetime_from_timestamp(ts: int) -> datetime:
    """Converts the specified unixtime timestamp to the python date/time type.

    :param ts: Timestamp for conversion.
    :return: Date/time from timestamp.

    """
    return datetime.fromtimestamp(ts, tz=UTC)


# Claims of specification JWT


def set_issuer(payload: dict[str, Any], value: str, claim: str = "iss"):
    """Sets the issuer of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.1

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_subject(payload: dict[str, Any], value: str, claim: str = "sub"):
    """Sets the subject of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.2

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_audience(payload: dict[str, Any], value: str, claim: str = "aud"):
    """Sets the audience of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.3

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_expiration(
    payload: dict[str, Any],
    now: datetime,
    lifetime: timedelta,
    claim: str = "exp",
):
    """Sets the expiration time of a token.

    https://tools.ietf.org/html/rfc7519#section-4.1.4

    :param payload: Token payload.
    :param now: Current date/time value to set the claim.
    :param lifetime: The value of the token lifetime to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = datetime_to_timestamp(now + lifetime)


def set_not_before(
    payload: dict[str, Any],
    activation_time: datetime,
    claim: str = "nbf",
):
    """Sets the time `Not before`.

    https://tools.ietf.org/html/rfc7519#section-4.1.5

    :param payload: Token payload.
    :param activation_time: Date/time value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = datetime_to_timestamp(activation_time)


def set_issued_at(payload: dict[str, Any], now: datetime, claim: str = "iat"):
    """Sets the time of token issuance.

    https://tools.ietf.org/html/rfc7519#section-4.1.6

    :param payload: Token payload.
    :param now: Current date/time value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = datetime_to_timestamp(now)


def set_token_id(
    payload: dict[str, Any],
    value: Optional[str] = None,
    claim: str = "jti",
):
    """Sets the token identifier.

    https://tools.ietf.org/html/rfc7519#section-4.1.7

    :param payload: Token payload.
    :param value: Value to set the claim. Default UUID4.
    :param claim: Claim short name (key).

    """
    payload[claim] = value if value else uuid4().hex


def check_expiration(
    payload: dict[str, Any],
    now: datetime,
    claim: str = "exp",
):
    """Checks if the timestamp value has passed in the given claim
    (since the given time value in `now`).
    If yes, it causes a `InvalidTokenError` with an error message.

    :param payload: Token payload.
    :param now: Current date/time value to check the claim.
    :param claim: Claim short name (key).

    :raise InvalidJWTTokenError:
    :raise ValidationJWTTokenError:

    """
    try:
        expiration_value = payload[claim]
    except KeyError:
        raise InvalidJWTTokenError(f"Token has no `{claim}` claim")

    expiration = datetime_from_timestamp(expiration_value)
    if now >= expiration:
        raise ValidationJWTTokenError(f"Token `{claim}` claim has expired")


def check_not_before(
    payload: dict[str, Any],
    now: datetime,
    claim: str = "nbf",
):
    """Checks if the timestamp value has reached the specified claim
    (the moment of the specified `now` time value).
    If so, raises a `InvalidTokenError` with an error message.

    :param payload: Token payload.
    :param now: Current date/time value to check the claim.
    :param claim: Claim short name (key).

    :raise InvalidJWTTokenError:
    :raise ValidationJWTTokenError:

    """
    try:
        not_before_value = payload[claim]
    except KeyError:
        raise InvalidJWTTokenError(f"Token has no `{claim}` claim")

    not_before_value = datetime_from_timestamp(not_before_value)
    if now < not_before_value:
        raise ValidationJWTTokenError(f"Token `{claim}` claim not reached")


# Custom JWT claims


def set_token_type(payload: dict[str, Any], value: str, claim: str = "jtt"):
    """Sets the token type claim.

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_permission_scopes(
    payload: dict[str, Any],
    scopes: str | Iterable[str],
    claim: str = "scp",
):
    """Sets the claim of the permissions scopes.

    :param payload: Token payload.
    :param scopes: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = ",".join(iterable_str_arg(scopes))


# Token handlers


def decode_jwt_token(
    token: str,
    algorithms: Iterable[str],
    signing_key: str,
    verify_key: str,
    verify: bool = True,
    decode_options: dict[bool | list[str]] = None,
    verify_values: dict[str, str | Iterable[str] | re.Pattern | None] = None,
    leeway: int | float = 0,
    jwk_url: Optional[str] = None,
) -> dict[str, Any]:
    """Performs validation of the provided token and returns
    its payload dictionary.

    :param token: JWT token.
    :param algorithms: Supported algorithms.
    :param signing_key: Signing key if used symmetric algorithm.
    :param verify_key: Public key if used asymmetric algorithm.
    :param verify: Verify token.
    :param decode_options: Options argument for `pyjwt.decode`.
        Example:
        `{
            "require": ["iss", "exp", "jti"],
            "verify_iss": False,
            "verify_exp": True,
            "verify_jti": False,
        }`
    :param verify_values: Values to be verified.
    :param leeway: Leeway time in seconds.
    :param jwk_url: URL for the PyJWK client.
        It gets the validation key for the specified jwt-token.
    :return: Decoded JWT token payload.

    :raise JWTBackendError: In the case of incorrectly passed
        parameters and arguments.
    :raise ValidationJWTTokenError: In case the token has an invalid signature
        or invalid payload data.

    """

    header = get_unverified_header(token)
    if header["alg"].startswith("HS"):
        verifying_key = signing_key
    elif jwk_url:
        jwks_client = PyJWKClient(jwk_url)
        verifying_key = jwks_client.get_signing_key_from_jwt(token).key
    else:
        verifying_key = verify_key

    verify_values = {} if verify_values is None else verify_values.copy()
    options = {"verify_signature": verify, **decode_options}
    params = {"leeway": leeway}

    # Provide values to check in the pyjwt module
    if "verify_iss" in options and options["verify_iss"] is True:
        # Issuer - One
        params["issuer"] = verify_values.pop("iss")
    if "verify_aud" in options and options["verify_aud"] is True:
        # Audience - One or Many
        if isinstance(verify_values.get("aud"), (str, list)):
            params["audience"] = verify_values.pop("aud")
        else:
            # todo: will be validate in `_extra_validation`
            options["verify_aud"] = False

    try:
        token_data = decode_complete(
            jwt=token,
            key=verifying_key,
            algorithms=algorithms,
            options=options,
            verify=verify,
            **params,
        )
    except InvalidAlgorithmError as error:
        raise JWTBackendError("Invalid algorithm specified") from error
    except InvalidTokenError:
        raise ValidationJWTTokenError("Token is invalid or expired")

    if verify:
        _extra_validation(token_data, verify_values)

    return token_data["payload"]


def _extra_validation(
    decoded_token: dict[str, dict[str, Any] | str],
    verify_values: dict[str, str | Iterable[str] | re.Pattern] = None,
):
    """Performs an additional check that is not performed by the pyjwt module.

    The check is performed by comparing the data provided for the check with
    the data included in the token payload with the same keys.

    :param decoded_token: Dict returned from `pyjwt.decode_complete`. Must
        contain a `header' and `payload'.
    :param verify_values: Dict with data to be validated. The keys must match
        the dictionary fields with the token payload.
    :raise ValidationJWTTokenError: In case the token has an invalid
        payload or header data.

    """
    header = decoded_token["header"]
    verify_header_values = verify_values.pop("header", [])
    payload = decoded_token["payload"]

    try:
        for key, value in verify_header_values:
            token_value = header.get(key)
            if isinstance(value, (str, int)):
                assert value == token_value
            elif isinstance(value, Iterable):
                assert token_value in value
            elif isinstance(value, re.Pattern):
                assert re.match(value, token_value)
    except AssertionError as error:
        raise ValidationJWTTokenError("Invalid token header") from error

    try:
        for key, value in verify_values:
            token_value = payload.get(key)
            if isinstance(value, (str, int)):
                assert value != token_value
            elif isinstance(value, Iterable):
                assert token_value not in value
            elif isinstance(value, re.Pattern):
                assert value.match(token_value)
    except AssertionError as error:
        raise ValidationJWTTokenError("Invalid token payload") from error




import gzip

from drakaina.middleware.base import BaseMiddleware


def gzip_str_to_file(raw_text, dest_file):
    with gzip.GzipFile(filename="", mode="w", fileobj=dest_file) as gz:
        gz.write(raw_text)


def gunzip_file(source_file):
    with gzip.GzipFile(filename="", mode="r", fileobj=source_file) as gz:
        return gz.read()


class GZIPMiddleware(BaseMiddleware):
    """
    todo
    """



from logging import Logger
from typing import Optional

from drakaina._types import ASGIApplication
from drakaina._types import ASGIReceive
from drakaina._types import ASGIScope
from drakaina._types import ASGISend
from drakaina._types import WSGIApplication
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse
from drakaina.middleware.base import BaseMiddleware
from drakaina.rpc_protocols import BaseRPCProtocol


class ExceptionMiddleware(BaseMiddleware):
    """The middleware for handling unhandled exceptions in the application
    according to the RPC protocol.

    """

    __slots__ = ("handler", "_rpc_content_type", "_logger")

    def __init__(
        self,
        app: ASGIApplication | WSGIApplication,
        handler: BaseRPCProtocol,
        logger: Optional[Logger] = None,
        is_async: bool = False,
    ):
        super().__init__(app=app, is_async=is_async)

        self.handler = handler
        self._rpc_content_type = self.handler.content_type
        self._logger = logger

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        try:
            return self.app(environ, start_response)
        except Exception as error:
            if self._logger is not None:
                self._logger.exception(error)
            response_body = self.handler.get_raw_error(error)
            response_headers = [
                ("Content-Type", self._rpc_content_type),
                ("Content-Length", str(len(response_body))),
            ]
            start_response("200 OK", response_headers)
            return (response_body,)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        try:
            await self.app(scope, receive, send)
        except Exception as error:
            if self._logger is not None:
                self._logger.exception(error)



from typing import Any
from typing import Optional
from typing import Type
from typing import Union

from drakaina import rpc_registry
from drakaina.exceptions import InternalServerError
from drakaina.exceptions import RPCError
from drakaina.registries import RPCRegistry
from drakaina.serializers import BaseSerializer

__all__ = ("BaseRPCProtocol",)


class BaseRPCProtocol:
    """Base class for representing the remote procedure call (RPC) protocol.

    :param registry:
        Registry of remote procedures.
        Default: `drakaina.registries.rpc_registry` (generic module instance)
    :type registry: RPCRegistry
    :param serializer:
        Serializer object.
    :type serializer: BaseSerializer

    """

    __slots__ = ("registry", "serializer", "_errors_map")

    def __init__(
        self,
        registry: Optional[RPCRegistry] = None,
        serializer: Optional[BaseSerializer] = None,
    ):
        self.registry = registry if registry is not None else rpc_registry
        self.serializer = serializer

        self._errors_map = {Exception: RPCError}

    def handle_raw_request(
        self,
        raw_data: bytes,
        request: Optional[Any] = None,
    ) -> bytes:
        """Accepts raw data, serializes, processes the RPC request,
        and returns the deserialized result.

        :param raw_data:
            Raw request data.
        :type raw_data:
        :param request:
            Request object or context data. Can be provided to
            a remote procedure.
        :type request: Any
        :return:
            Deserialized RPC response data.

        """
        try:
            parsed_data = self.serializer.deserialize(raw_data)
        except Exception as exc:
            return self.get_raw_error(exc)

        response_data = self.handle(parsed_data, request=request)
        if response_data is None:
            return b""

        try:
            return self.serializer.serialize(response_data)
        except Exception as exc:
            return self.get_raw_error(exc)

    def get_raw_error(
        self,
        error: Union[RPCError, Exception, Type[Exception]],
    ) -> bytes:
        """Returns the deserialized error object.

        :param error:
            The instance or class of the error.
        :type error: RPCError | Type[RPCError] | Exception
        :return:
            Raw error data.

        """

        if isinstance(error, type) and issubclass(error, Exception):
            error = error()

        rpc_error_class = self._errors_map.get(type(error), self.default_error)
        rpc_error = rpc_error_class(str(error))

        return self.serializer.serialize(rpc_error.as_dict())

    def handle(self, rpc_request: Any, request: Optional[Any] = None) -> Any:
        """Handles a procedure call.

        :param rpc_request:
            RPC request in protocol format.
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure. By default, None will be passed.
        :return:
            Returns the result in protocol format.

        """
        raise NotImplementedError(
            "You must implement the `handle` method in the child class",
        )

    @property
    def default_error(self):
        return InternalServerError

    @property
    def content_type(self) -> str:
        return self.serializer.content_type

    def smd_scheme(self) -> bytes:
        raise NotImplementedError

    def openrpc_scheme(self) -> bytes:
        """Implementation of OpenRPC specification

        https://open-rpc.org/getting-started
        https://spec.open-rpc.org/#introduction
        """
        raise NotImplementedError

    def openapi_scheme(self) -> bytes:
        """Implementation of OpenAPI specification

        https://github.com/OAI/OpenAPI-Specification
        """
        raise NotImplementedError



import typing

import aiohttp

from .base import BaseJsonRpcClient
from .. import errors, utils


__all__ = (
    'JsonRpcClient',
)


class JsonRpcClient(BaseJsonRpcClient):
    url: str
    session: typing.Optional[aiohttp.ClientSession]
    request_kwargs: dict
    _session_is_outer: bool

    def __init__(self,
                 url: str, *,
                 session: typing.Optional[aiohttp.ClientSession] = None,
                 **request_kwargs) -> None:
        self.url = url
        self.session = session
        self.request_kwargs = request_kwargs
        self._session_is_outer = session is not None  # We don't close an outer session.

    async def connect(self) -> None:
        if self.session is None:
            self.session = aiohttp.ClientSession(json_serialize=self.json_serialize, **self.request_kwargs)

    async def disconnect(self) -> None:
        if self.session is not None and not self._session_is_outer:
            await self.session.close()

    async def send_json(self,
                        data: typing.Any, *,
                        without_response: bool = False,
                        **kwargs) -> typing.Tuple[typing.Any, typing.Optional[dict]]:
        assert self.session is not None

        http_response = await self.session.post(self.url, json=data, **kwargs)

        try:
            http_response.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise errors.ServerError(f'Server responded with code {http_response.status}.') from e

        try:
            json_response = await http_response.json()
        except aiohttp.ContentTypeError as e:
            raise errors.ParseError(utils.get_exc_message(e)) from e

        if without_response:
            return None, None

        return json_response, {'http_response': http_response}



from __future__ import annotations

from drakaina._types import ASGIApplication
from drakaina._types import ASGIReceive
from drakaina._types import ASGIScope
from drakaina._types import ASGISend
from drakaina._types import WSGIApplication
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse


__all__ = ("BaseMiddleware",)


class BaseMiddleware:
    """Base class for middleware.

    It is a simple middleware for WSGI and ASGI, where the protocol
    will be chosen depending on the `is_async` parameter.

    :param app:
    :type app: WSGIApplication | ASGIApplication
    """

    __slots__ = ("app", "is_async", "__call__")

    def __init__(
        self,
        app: ASGIApplication | WSGIApplication,
        is_async: bool = False,
    ):
        self.app = app
        self.is_async = is_async
        if is_async:
            self.__call__ = self.__asgi_call__
        else:
            self.__call__ = self.__wsgi_call__

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        await self.app(scope, receive, send)



from collections.abc import Callable
from inspect import getfullargspec
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple

from drakaina.utils import unwrap_func

__all__ = (
    "is_rpc_procedure",
    "RPCRegistry",
    "RPC_REGISTRY",
    "RPC_NAME",
    "RPC_REGISTERED",
    "RPC_PROVIDE_REQUEST",
    "RPC_META",
)

# Reserved procedure argument names
RESERVED_KWARGS = ("self", "request")

# RPC procedure fields
RPC_REGISTRY = "__rpc_registry"
RPC_NAME = "__rpc_name"
RPC_REGISTERED = "__rpc_procedure"
RPC_PROVIDE_REQUEST = "__rpc_provide_request"
RPC_META = "__rpc_metadata"


def is_rpc_procedure(func: Callable) -> bool:
    return hasattr(unwrap_func(func), RPC_REGISTERED)


class RPCRegistry:
    """Registry of remote procedures"""

    _remote_procedures: Dict[str, Callable[..., Any]]

    def __init__(self):
        self._remote_procedures = {}

    def register_procedure(
        self,
        procedure: Callable[..., Any],
        name: Optional[str] = None,
        provide_request: Optional[bool] = None,
        metadata: Optional[dict] = None,
    ):
        """Register a function as a remote procedure.

        :param procedure:
            Registered procedure.
        :type procedure: Callable
        :param name:
            Procedure name. Default as function name.
        :type name: str
        :param provide_request:
            If `True`, then the request object or context can be supplied to
            the procedure as a `request` argument.
        :type provide_request: bool
        :param metadata:
            Metadata that can be processed by middleware.
        :type metadata: dict

        """
        assert callable(procedure)

        procedure_name = procedure.__name__ if name is None else name
        # The loop is for applying attributes to wrapped functions.
        # If the provided function is not wrapped, the attributes will be
        # applied once
        _func = procedure
        while _func is not None:
            setattr(_func, RPC_REGISTRY, self)
            setattr(_func, RPC_NAME, procedure_name)
            setattr(_func, RPC_REGISTERED, True)
            setattr(_func, RPC_PROVIDE_REQUEST, provide_request)
            setattr(_func, RPC_META, metadata or {})
            # If it's a wrapped function
            _func = getattr(_func, "__wrapped__", None)

        procedure.__rpc_args = [
            a
            for a in getfullargspec(procedure).args
            if a not in RESERVED_KWARGS
        ]

        self._remote_procedures[procedure_name] = procedure

    def replace(self, procedure: Callable, new_procedure: Callable):
        procedure_name = getattr(procedure, RPC_NAME, None)
        self._remote_procedures[procedure_name] = new_procedure

    def __getitem__(self, key: str) -> Optional[Callable]:
        return self._remote_procedures.get(key)

    def __setitem__(self, key: str, value: Callable):
        self.register_procedure(procedure=value, name=key)

    def __delitem__(self, key: str):
        del self._remote_procedures[key]

    def __len__(self) -> int:
        return len(self._remote_procedures)

    def __iter__(self) -> Iterator[str]:
        yield self._remote_procedures

    def get(
        self,
        key: str,
        default: Optional[Callable] = None,
    ) -> Optional[Callable]:
        return self._remote_procedures.get(key, default)

    def items(self) -> Iterator[Tuple[str, Callable]]:
        yield self._remote_procedures.items()



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Protocol

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpRequest
from django.http import HttpResponse
from django.utils.module_loading import import_string

from drakaina import ENV_AUTH_PAYLOAD
from drakaina import ENV_AUTH_SCOPES
from drakaina import ENV_IS_AUTHENTICATED
from drakaina import ENV_USER_ID
from drakaina.contrib import django as default_settings
from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.utils import decode_jwt_token
from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import ForbiddenError
from drakaina.rpc_protocols import BaseRPCProtocol

__all__ = ("JWTAuthenticationMiddleware",)

UserModel = get_user_model()
ARG_PREFIX = "DRAKAINA_"


class DjangoRPCView(Protocol):
    handler: BaseRPCProtocol

    def __call__(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        ...


def get_settings_arg(settings_attribute: str) -> Any:
    return getattr(
        settings,
        ARG_PREFIX + settings_attribute,
        getattr(default_settings, settings_attribute, None),
    )


class JWTAuthenticationMiddleware:
    """The middleware supporting JWT tokens.

    This middleware uses with `RPCView`.
    For configure this middleware define in settings module

    """

    csrf_exempt = True

    def __init__(self, view: DjangoRPCView):
        assert hasattr(
            getattr(view, "cls", None),
            "handler",
        ), "The view must contain the BaseRPCProtocol implementation instance."

        self._rpc_view = view
        self._rpc_handler = view.cls.handler  # noqa

        self.credentials_required = get_settings_arg("CREDENTIALS_REQUIRED")
        self.prefix = get_settings_arg("JWT_PREFIX")
        self.cookie_key = get_settings_arg("JWT_COOKIE_KEY")
        self.use_cookies = get_settings_arg("JWT_USE_COOKIES")
        self.algorithms = get_settings_arg("JWT_ALGORITHMS")
        self.secret_key = get_settings_arg("JWT_SECRET_KEY")
        self.public_key = get_settings_arg("JWT_PUBLIC_KEY")
        self.decode_options = get_settings_arg("JWT_DECODE_OPTIONS")
        self.verify_values = get_settings_arg("JWT_VERIFY_VALUES")
        self.leeway = get_settings_arg("JWT_LEEWAY")
        self.user_id_field = get_settings_arg("JWT_USER_ID_FIELD")

        self.get_token = self._get_token
        token_getter = get_settings_arg("JWT_TOKEN_GETTER")
        if token_getter:
            self.get_token = import_string(token_getter)

        self.get_user = self._get_user
        user_getter = get_settings_arg("JWT_USER_GETTER")
        if user_getter:
            self.get_user = import_string(user_getter)

        self.get_scopes = None
        scopes_getter = get_settings_arg("JWT_SCOPES_GETTER")
        if scopes_getter:
            self.get_scopes = import_string(scopes_getter)

        self.token_is_revoked = None
        revoke_checker = get_settings_arg("JWT_REVOKE_CHECKER")
        if revoke_checker:
            self.token_is_revoked = import_string(revoke_checker)

    def __call__(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        try:
            token = self.get_token(request)
        except Exception as error:
            return self._error_response(error)

        if not token and self.credentials_required:
            return self._error_response(
                AuthenticationFailedError("Credential required"),
            )

        if token is not None:
            payload = decode_jwt_token(
                token,
                algorithms=self.algorithms,
                signing_key=self.secret_key,
                verify_key=self.public_key,
                verify=True,
                decode_options=self.decode_options,
                verify_values=self.verify_values,
                leeway=self.leeway,
            )

            if callable(self.token_is_revoked):
                if self.token_is_revoked(request, payload):
                    return self._error_response(
                        ForbiddenError("Token is revoked"),
                    )

            setattr(request, ENV_USER_ID, payload[self.user_id_field])
            setattr(request, ENV_IS_AUTHENTICATED, True)
            setattr(request, ENV_AUTH_PAYLOAD, payload)

            if callable(self.get_user):
                request.user = self.get_user(request, payload)
            if callable(self.get_scopes):
                setattr(
                    request,
                    ENV_AUTH_SCOPES,
                    self.get_scopes(request, payload),
                )

        return self._rpc_view(request, *args, **kwargs)

    def _get_token(self, request: HttpRequest) -> Optional[str]:
        if self.use_cookies:
            return request.COOKIES.get(self.cookie_key)
        else:
            auth_header = request.META.get("Authorization")
            if auth_header is None:
                return None

            try:
                parts = auth_header.strip().split(" ")
            except ValueError:
                raise AuthenticationFailedError(
                    "Invalid `Authorization` header",
                )

            if len(parts) == 0 or parts[0] not in self.prefix:
                return None

            if len(parts) != 2:
                raise InvalidJWTTokenError(
                    "The `Authorization` header must contain two values "
                    "separated by a space",
                )

            return parts[1]

    def _get_user(
        self,
        request: HttpRequest,
        payload: dict[str, Any],
    ) -> Optional[UserModel]:
        user_id = payload.get(self.user_id_field)
        if user_id is None:
            return None  # todo: try returns request.user
        return UserModel.objects.get(pk=user_id)

    def _error_response(self, error: Exception) -> HttpResponse:
        return HttpResponse(
            content=self._rpc_handler.get_raw_error(error),
            content_type=self._rpc_handler.content_type,
        )



""" Utility functions for package."""
from __future__ import annotations

from functools import partial
from typing import Callable
from typing import Iterable


def iterable_str_arg(s: str | Iterable[str]) -> list[str]:
    if isinstance(s, str):
        s = s.replace(",", " ").replace("  ", " ").split()
    return list(map(str, s))


def unwrap_func(func: Callable) -> Callable:
    _func = func
    while hasattr(_func, "__wrapped__") or isinstance(_func, partial):
        _func = getattr(_func, "__wrapped__") or _func.func
    return _func


def get_cookies(cookie_str: str) -> dict[str, str]:
    if cookie_str in None:
        return {}

    _cookies = {}
    for cookie_item in cookie_str.split(";"):
        if "=" not in cookie_item or "DELETED" in cookie_item:
            continue
        name, value = cookie_item.strip().split("=", 1)
        _cookies[name] = value

    return _cookies



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drakaina',
 'drakaina.client',
 'drakaina.contrib',
 'drakaina.contrib.django',
 'drakaina.contrib.jwt',
 'drakaina.middleware',
 'drakaina.middleware.openapi',
 'drakaina.rpc_protocols']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.5.0,<5.0.0']

extras_require = \
{'jwt': ['pyjwt>=2.6.0,<3.0.0'],
 'msgpack': ['msgpack>=1.0.4,<2.0.0'],
 'orjson': ['orjson>=3.8.5,<4.0.0'],
 'tests': ['pyjwt>=2.6.0,<3.0.0'],
 'ujson': ['ujson>=5.7.0,<6.0.0']}

setup_kwargs = {
    'name': 'drakaina',
    'version': '0.6.7',
    'description': 'Module for simple RPC service implementation',
    'long_description': '# drakaina\n\n![Drakaina](content/drakaina.png "Drakaina"){width=200px height=205px}\n\n[![image](https://img.shields.io/pypi/v/drakaina.svg)](https://pypi.python.org/pypi/drakaina)\n[![image](https://img.shields.io/pypi/l/drakaina.svg)](https://pypi.python.org/pypi/drakaina)\n[![image](https://img.shields.io/pypi/pyversions/drakaina.svg)](https://pypi.python.org/pypi/drakaina)\n[![Ruff](https://img.shields.io/endpoint?url=https://raw.githubusercontent.com/charliermarsh/ruff/main/assets/badge/v1.json)](https://github.com/charliermarsh/ruff)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-black.svg)](https://github.com/psf/black)\n[![libera manifesto](https://img.shields.io/badge/libera-manifesto-lightgrey.svg)](https://liberamanifesto.com)\n\n❗ WIP\n\nModule for simple RPC service implementation\n\n\n## Quickstart\n\nDrakaina may be installed via `pip` and requires Python 3.7 or higher :\n\n```shell\npip install drakaina\n```\n\nA minimal Drakaina example is:\n\n```python\nfrom drakaina import remote_procedure\nfrom drakaina.wsgi import WSGIHandler\n\n\n@remote_procedure\ndef my_method():\n    return "Hello Bro! ✋️"\n\n\n@remote_procedure(name="something.get")\ndef get_some_string():\n    return "You called `something.get`."\n\n\n@remote_procedure(provide_request=True)\ndef do_something_with_environ(request):\n    return f"You called `do_something_with_environ`. Request: {request}."\n\n\n@remote_procedure()\ndef tell_the_middleware_something():\n    return "You called `tell_the_middleware_something`. It has a some extra conditions."\n\n\nasync def asynchronous_procedure():\n    await asyncio.sleep(5)\n    return "Ding-Dong 🔔!"\n\n\n"""\n>>> JsonRPCv2().handle({"jsonrpc": "2.0", "method": "my_method", "id": 1})\nor define WSGI application\n"""\napp = WSGIHandler(route="/jrpc")\n```\n\nDrakaina may be ran with any WSGI-compliant server,\nsuch as [Gunicorn](http://gunicorn.org).\n\n```shell\ngunicorn main:app\n```\n\n\n## Features\n\n- WSGI protocol implementation\n  - Implemented CORS middleware\n  - Compatible with simple middlewares for others wsgi-frameworks,\n    like as [Werkzeug](https://palletsprojects.com/p/werkzeug/),\n    [Flask](https://palletsprojects.com/p/flask/)\n\n\n# Documentation\n\n\n## Installation\n\n```shell\npip install drakaina\n```\n\n\n## Middlewares\n\n\n### CORS\n\n\n### JWT\n\nDrakaina may be installed via `pip` and requires Python 3.7 or higher :\n\n```shell\npip install drakaina[jwt]\n```\n\nA minimal Drakaina example is:\n\n```python\nfrom functools import partial\nfrom drakaina import ENV_IS_AUTHENTICATED\nfrom drakaina import ENV_USER_ID\nfrom drakaina import remote_procedure\nfrom drakaina import check_permissions\nfrom drakaina import login_required\nfrom drakaina import match_any\nfrom drakaina.contrib.jwt.middleware import JWTAuthenticationMiddleware\nfrom drakaina.wsgi import WSGIHandler\n\n\n@login_required\n@remote_procedure(provide_request=True)\ndef my_method(request):\n    assert request[ENV_IS_AUTHENTICATED]\n    return f"Hello Bro ✋! Your ID={request[ENV_USER_ID]}"\n\n\n@check_permissions(\n    scopes=["user_read", "app/user:admin", "username:johndoe"],\n    comparator=match_any,\n)\n@remote_procedure\ndef my_method():\n    return "Hello Bro! ✋️"\n\n\n\n\napp = WSGIHandler(\n    middlewares=[\n        partial(\n            JWTAuthenticationMiddleware,\n            secret_phrase="_secret_",\n            credentials_required=True,\n        )\n    ]\n)\n```\n\n\n### Using with Django\n\nCreate file `rpc_views.py` in your django application.\nDefine function and wrap it `remote_procedure` decorator:\n\n```python\nfrom drakaina import remote_procedure\n\n@remote_procedure\ndef my_method():\n    return "Hello, Django Bro! ✋"\n```\n\nAdd `RPCView` class to urlpatterns. The `as_view` method\nmust accept the `autodiscover` argument as the name of\nthe remote procedure files.\n\n```python\nfrom django.urls import path\nfrom drakaina.contrib.django.views import RPCView\n\nurlpatterns = [\n    ...,\n    path("api/", RPCView.as_view(autodiscover="rpc_views")),\n]\n```\n\n\n### JWT Authentication in your Django project\n\nWrap an instance of `RPCView` with the `JWTAuthenticationMiddleware`.\n\n```python\nfrom django.urls import path\nfrom drakaina.contrib.django import RPCView, JWTAuthenticationMiddleware\n\nurlpatterns = [\n    ...,\n    path("api/", JWTAuthenticationMiddleware(\n        RPCView.as_view(autodiscover="rpc_views")\n    )),\n]\n```\n\nDefine the parameters in the `settings.py` file.\n\n```python\n...\n\nDRAKAINA_JWT_SECRET_KEY = SECRET_KEY\n\n...\n```\n\n\n## License\n\nApache License 2.0\n\n## Artwork\n\n"[drakaina.png](content/drakaina.png)" by Korolko Anastasia is licensed under\n<a rel="license" href="http://creativecommons.org/licenses/by-sa/4.0/"><img alt="License Creative Commons" style="border-width:0" src="https://i.creativecommons.org/l/by-sa/4.0/80x15.png" /></a> ([CC BY-SA 4.0](http://creativecommons.org/licenses/by-sa/4.0/)).\n',
    'author': 'Aleksey Terentyev',
    'author_email': 'terentyev.a@pm.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/tau_lex/drakaina',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
The Client class, used for accessing JSONRPCTCP Server instances. It
shouldn't need to be instantiated directly, instead, use the
connect function:

conn = connect('localhost', 8001)
result = conn.method(param1, param2)
result = conn.tree.method(keyword=arg)
"""
import json
import logging
import socket
import uuid

from . import config

EncryptionMissing, ProtocolError = Exception, Exception
history = object
logger = logging.getLogger(__name__)


class Client(object):
    """
    This is the JSON RPC client class, which translates attributes into
    function calls and request / response translations, and organizes
    batches, notifications, etc.
    """

    _requests = None
    _request = None
    _response = None

    def __init__(self, addr, **kwargs):
        self._addr = addr
        self._requests = []
        self.__batch = kwargs.get("batch", None)
        self._key = kwargs.get("key", None)
        if self._key and not config.crypt:
            raise EncryptionMissing("No encryption library found.")

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError("Methods that start with _ are not allowed.")
        req_id = "%s" % uuid.uuid4()
        request = ClientRequest(self, namespace=key, req_id=req_id)
        self._requests.append(request)
        return request

    @property
    def _notification(self):
        """
        Returns a specialized version of the ClientRequest object,
        which is prepped for notification.
        """
        request = ClientRequest(self, notify=True, req_id=None)
        self._requests.append(request)
        return request

    def _batch(self):
        """
        Returns a specialized version of the Client class, prepped for
        a series of calls which will only be sent when the Client is
        __call__()ed.
        """
        return Client(self._addr, batch=True)

    def _is_batch(self):
        """Checks whether the batch flag is set."""
        return self.__batch is True

    def __call__(self):
        assert len(self._requests) > 0
        requests = []
        for i in range(len(self._requests)):
            request = self._requests.pop(0)
            requests.append(request._request())
        if not self._is_batch():
            result = self._call_single(requests[0])
        else:
            result = self._call_batch(requests)
        self._requests = []
        return result

    def _call_single(self, request):
        """
        Processes a single request, and returns the response.
        """
        self._request = request
        message = json.dumps(request)
        notify = False
        if "id" not in request:
            notify = True
        response_text = self._send_and_receive(message, notify=notify)
        response = self._parse_response(response_text)
        if not response:
            return response
        self._response = response
        validate_response(response)
        return response["result"]

    def _call_batch(self, requests):
        """
        Processes a batch, and returns a generator to iterate over the
        response results.
        """
        ids = []
        for request in requests:
            if request.has_key("id"):
                ids.append(request["id"])
        self._request = requests
        message = json.dumps(requests)
        notify = False
        if len(ids) == 0:
            notify = True
        response_text = self._send_and_receive(
            message, batch=True, notify=notify
        )
        responses = self._parse_response(response_text)
        if responses is None:
            responses = []
        assert type(responses) is list
        return BatchResponses(responses, ids)

    def _send_and_receive(self, message, batch=False, notify=False):
        """
        Handles the socket connection, sends the JSON request, and
        (if not a notification) retrieves the response and decodes the
        JSON text.
        """
        # Starting with a clean history
        history.request = message
        logger.debug("CLIENT | REQUEST: %s" % message)
        if self._key:
            crypt = config.crypt.new(self._key)
            length = config.crypt_chunk_size
            pad_length = length - (len(message) % length)
            message = crypt.encrypt("%s%s" % (message, " " * pad_length))
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(config.timeout)
        sock.connect(self._addr)
        if sys.version_info[0] == 2:
            sock.send(message)
        else:
            sock.send(message.encode("utf-8"))

        responselist = []
        if notify:
            # single notification, we don't need a response.
            sock.close()
        else:
            while True:
                try:
                    data = sock.recv(config.buffer)
                except socket.timeout:
                    break
                if not data:
                    break
                responselist.append(data)
                if len(data) < config.buffer:
                    break
            sock.close()
        if sys.version_info[0] == 2:
            response = "".join(responselist)
        else:
            response = b"".join(responselist)
        if self._key:
            try:
                response = crypt.decrypt(response)
            except ValueError:
                # What exactly is an intuitive response to a poorly- or
                # not-encrypted response to an encrypted request?
                raise ProtocolError(-32700, "Response not encrypted properly.")
            # Should we do a preliminary json.loads here to verify that the
            # decryption succeeded?
        logger.debug("CLIENT | RESPONSE: %s" % response)
        history.response = response
        return response

    def _parse_response(self, response):
        if response == "":
            return None
        try:
            obj = json.loads(response)
        except ValueError:
            raise ProtocolError(-32700)
        if obj == dict() and "error" in obj:
            raise ProtocolError(
                obj.get("error").get("code"),
                obj.get("error").get("message"),
                obj.get("error").get("data", None),
            )
        return obj


class BatchResponses(object):
    """
    This is just a wrapper around the responses so you can
    iterate or retrieve by single id.
    """

    def __init__(self, responses, ids):
        self.responses = responses
        self.ids = ids
        response_by_id = {}
        for response in responses:
            response_id = response.get("id", None)
            response_by_id.setdefault(response_id, [])
            response_by_id[response_id].append(response)
        self._response_by_id = response_by_id

    def __iter__(self):
        for request_id in self.ids:
            yield self.get(request_id)

    def get(self, req_id):
        responses = self._response_by_id.get(req_id, None)
        if not responses:
            responses = self._response_by_id.get(None)
        if not responses or len(responses) == 0:
            raise KeyError(
                'Job "%s" does not exist or has already be retrieved.' % req_id
            )
        response = responses.pop(0)
        validate_response(response)
        return response["result"]


class ClientRequest(object):
    """
    This is the class that holds all of the namespaced methods,
    as well as whether or not it is a notification. When it is
    finally called, it parses the arguments and passes it to
    the parent Client.
    """

    def __init__(self, client, namespace="", notify=False, req_id=None):
        self._client = client
        self._namespace = namespace
        self._notification = notify
        self._req_id = req_id
        self._params = None

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError
        if self._namespace:
            self._namespace += "."
        self._namespace += key
        return self

    def __call__(self, *args, **kwargs):
        if not (len(args) == 0 or len(kwargs) == 0):
            raise ValueError(
                "JSON spec allows positional arguments OR "
                + "keyword arguments, not both."
            )
        params = list(args)
        if len(kwargs) > 0:
            params = kwargs
        return self._call_server(params)

    def _call_server(self, params):
        """
        Forms a valid jsonrpc query, and passes it on to the parent
        Client, returning the response.
        """
        self._params = params
        if not self._client._is_batch():
            return self._client()
        # Add batch logic here

    def _request(self):
        request = {"jsonrpc": "2.0", "method": self._namespace}
        if self._params:
            request["params"] = self._params
        if not self._notification:
            request["id"] = self._req_id
        return request


def connect(host, port, key=None):
    """
    This is a wrapper function for the Client class.
    """
    client = Client((host, port), key=key)
    return client


def validate_response(response):
    """
    Parses the returned JSON object, verifies that it follows
    the JSON-RPC spec, and checks for errors, raising exceptions
    as necessary.
    """
    jsonrpc = "jsonrpc" in response
    response_id = "id" in response
    result = "result" in response
    error = "error" in response
    if not jsonrpc or not response_id or (not result and not error):
        raise Exception("Server returned invalid response.")
    if error:
        raise ProtocolError(
            response["error"]["code"], response["error"]["message"]
        )


def test_client():
    """
    This is the test client to be run against the test_server in
    the server module.
    """
    conn = connect("localhost", 8080)
    value = "Testing!"
    result = conn.echo(value)
    assert result == value
    print("Single test completed.")

    result = conn._notification.echo(message="No response!")
    assert result is None
    print("Notify test completed.")

    batch = conn._batch()
    batch.tree.echo(message="First!")
    batch._notification.echo("Skip!")
    batch.tree.echo("Last!")
    results = []
    for i in batch():
        results.append(i)
    assert results == ["First!", "Last!"]
    print("Batch test completed.")

    result = conn.echo(message=5)
    assert result == 5
    print("Post-batch test completed.")

    try:
        conn.echo()
    except Exception as e:
        print("Bad call had necessary exception.")
        print(e.code, e.message)
    else:
        print("ERROR: Did not throw exception for bad call.")

    try:
        conn.foobar(5, 6)
    except Exception as e:
        print("Invalid method threw exception.")
        print(e.code, e.message)
    else:
        print("ERROR: Did not throw exception for bad method.")

    print("=============================")
    print("Tests completed successfully.")


if __name__ == "__main__":
    import sys
    import logging

    if "-v" in sys.argv:
        config.verbose = True
        logger.setLevel(logging.DEBUG)
        logger.addHandler(logging.StreamHandler())
    test_client()



"""Submodule with implementations support of JSON Web Tokens

Standard: https://www.rfc-editor.org/rfc/rfc7519
"""

__all__ = (
    "RESERVED_FIELDS",
    "SUPPORTED_ALGORITHMS",
)

RESERVED_FIELDS = ("iss", "sub", "aud", "exp", "nbf", "iat", "jti")
SUPPORTED_ALGORITHMS = ("HS256", "HS384", "HS512", "RS256", "RS384", "RS512")



from __future__ import annotations

from typing import Iterable
from typing import Optional

from drakaina._types import ASGIReceive
from drakaina._types import ASGIScope
from drakaina._types import ASGISend
from drakaina._types import WSGIApplication
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse
from drakaina.middleware.base import BaseMiddleware
from drakaina.utils import iterable_str_arg

__all__ = ("CORSMiddleware",)

DEFAULT_HEADERS = (
    "Accept, Accept-Encoding, Accept-Language, Authorization, "
    "Content-Language, Content-Type, DNT, Origin, User-Agent, X-Requested-With"
)
ALL_METHODS = "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
DEFAULT_METHODS = "GET, POST, OPTIONS"


class CORSMiddleware(BaseMiddleware):
    """Middleware for providing CORS headers and handling preflight requests.

    See: https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

    :param allow_origin:
        The "Access-Control-Allow-Origin" header.
    :type allow_origin: Union[str, Iterable[str]]
    :param allow_methods:
        The "Access-Control-Allow-Methods" header.
        Default: "GET, POST, OPTIONS".
    :type allow_methods: Union[str, Iterable[str]]
    :param allow_headers:
        The "Access-Control-Allow-Headers" header.
        Default: "Accept, Accept-Encoding, Authorization,
        Content-Type, DNT, Origin, User-Agent, X-Requested-With".
    :type allow_headers: Union[str, Iterable[str]]
    :param allow_credentials:
        The "Access-Control-Allow-Credentials" header. Default: None.
    :type allow_credentials: bool
    :param expose_headers:
        The "Access-Control-Expose-Headers" header. Default: None.
    :type expose_headers: Union[str, Iterable[str]]
    :param max_age:
        The "Access-Control-Max-Age" header. Default: 3600 sec. (1 hour).
    :type max_age: Union[int, str]

    """

    __slots__ = (
        "_cors_headers",
        "_preflight_cors_headers",
        "_allow_origin",
        "_allow_credentials",
        "_allow_headers",
    )

    def __init__(
        self,
        app: WSGIApplication,
        allow_origin: str | Iterable[str],
        allow_methods: Optional[str | Iterable[str]] = None,
        allow_headers: Optional[str | Iterable[str]] = None,
        allow_credentials: Optional[bool] = None,
        expose_headers: Optional[str | Iterable[str]] = None,
        max_age: int = 3600,
        **kwargs,
    ):
        super().__init__(app, **kwargs)

        self._allow_origin = ", ".join(iterable_str_arg(allow_origin))

        if allow_methods:
            if "*" in allow_methods:
                _allow_methods = ALL_METHODS
            else:
                _allow_methods = ", ".join(iterable_str_arg(allow_methods))
        else:
            _allow_methods = DEFAULT_METHODS

        if allow_headers:
            self._allow_headers = ", ".join(
                sorted(
                    set(iterable_str_arg(DEFAULT_HEADERS))
                    | set(iterable_str_arg(allow_headers)),
                ),
            )
        else:
            self._allow_headers = DEFAULT_HEADERS

        if expose_headers:
            _expose_headers = ", ".join(iterable_str_arg(expose_headers))
        else:
            _expose_headers = None

        self._allow_credentials = allow_credentials

        # Prepare CORS response headers
        self._cors_headers = []
        if self._allow_credentials:
            self._cors_headers.append(
                ("Access-Control-Allow-Credentials", "true"),
            )
        if _expose_headers:
            self._cors_headers.append(
                ("Access-Control-Expose-Headers", _expose_headers),
            )

        # Prepare pre-flight CORS response headers
        self._preflight_cors_headers = [
            ("Access-Control-Allow-Methods", _allow_methods),
            ("Access-Control-Max-Age", str(max_age)),
            ("Content-Length", "0"),
        ]
        if "*" not in self._allow_origin or self._allow_credentials:
            self._preflight_cors_headers.append(("Vary", "Origin"))

        if self._allow_credentials:
            self._preflight_cors_headers.append(
                ("Access-Control-Allow-Credentials", "true"),
            )

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        if environ.get("HTTP_ORIGIN"):
            if environ.get("REQUEST_METHOD") == "OPTIONS":
                return self.options(environ, start_response)
            return self.app(
                environ,
                self._start_response_with_cors(environ, start_response),
            )
        else:
            return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
        ...

    def options(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        """Response for OPTIONS request"""
        request_origin = environ["HTTP_ORIGIN"]
        request_headers = environ.get("HTTP_ACCESS_CONTROL_REQUEST_HEADERS")

        response_headers = self._preflight_cors_headers[:]

        if "*" in self._allow_origin or request_origin in self._allow_origin:
            if "*" not in self._allow_origin or self._allow_credentials:
                response_headers.append(
                    ("Access-Control-Allow-Origin", request_origin),
                )
            else:
                response_headers.append(("Access-Control-Allow-Origin", "*"))

        if "*" in self._allow_headers and request_headers is not None:
            response_headers.append(
                ("Access-Control-Allow-Headers", request_headers),
            )
        if "*" not in self._allow_headers:
            response_headers.append(
                ("Access-Control-Allow-Headers", self._allow_headers),
            )

        start_response("200 OK", response_headers)
        yield b""

    def _start_response_with_cors(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIStartResponse:
        """Wraps the start_response method, and includes the CORS header
        for the specified origin.
        """
        request_origin = environ["HTTP_ORIGIN"]

        cors_headers = self._cors_headers[:]

        if "*" in self._allow_origin:
            if "HTTP_COOKIE" in environ:
                cors_headers.append(
                    ("Access-Control-Allow-Origin", request_origin),
                )
                cors_headers.append(("Vary", "Origin"))
            else:
                cors_headers.append(("Access-Control-Allow-Origin", "*"))
        elif (
            "*" not in self._allow_origin
            and request_origin in self._allow_origin
        ):
            cors_headers.append(("Access-Control-Allow-Origin", request_origin))
            cors_headers.append(("Vary", "Origin"))

        def response_with_cors(status, headers, exc_info=None):
            headers.extend(cors_headers)
            return start_response(status, headers, exc_info)

        return response_with_cors



import asyncio
import json
from uuid import uuid4

from aiohttp import ClientSession

from validictory import validate, ValidationError, SchemaError
from .exc import (
    ParseError,
    InvalidRequest,
    InvalidParams,
    InternalError,
    InvalidResponse,
)


# Example with client side exceptions

REQ_JSONRPC20 = {
    "type": "object",
    "properties": {
        "jsonrpc": {"pattern": r"2\.0"},
        "method": {"type": "string"},
        "params": {"type": "any"},
        "id": {"type": "any"},
    },
}
RSP_JSONRPC20 = {
    "type": "object",
    "properties": {
        "jsonrpc": {"pattern": r"2\.0"},
        "result": {"type": "any"},
        "id": {"type": "any"},
    },
}
ERR_JSONRPC20 = {
    "type": "object",
    "properties": {
        "jsonrpc": {"pattern": r"2\.0"},
        "error": {
            "type": "object",
            "properties": {
                "code": {"type": "number"},
                "message": {"type": "string"},
            },
        },
        "id": {"type": "any"},
    },
}


class Response(object):

    __slots__ = ["id", "error", "result"]

    def __init__(self, id, result=None, error=None, **kw):
        self.id = id
        self.result = result
        self.error = error

    def __repr__(self):
        return "Response(id={rid}, result={res}, error={err}".format(
            rid=self.id, res=self.result, err=self.error
        )


class Client(object):
    def __init__(self, url, dumper=None, loop=None):
        self.url = url
        self.dumper = dumper
        if not loop:
            loop = asyncio.get_event_loop()
        if not self.dumper:
            self.dumper = json.dumps

        self.client = ClientSession(
            loop=loop, headers={"content-type": "application/json"}
        )

    def __del__(self):
        self.client.close()

    def __encode(self, method, params=None, id=None):
        try:
            data = self.dumper(
                {"jsonrpc": "2.0", "id": id, "method": method, "params": params}
            )
        except Exception as e:
            raise Exception("Can not encode: {}".format(e))

        return data

    async def call(self, method, params=None, id=None, schem=None):
        if not id:
            id = uuid4().hex
        try:
            resp = await self.client.post(
                self.url, data=self.__encode(method, params, id)
            )
        except Exception as err:
            raise Exception(err)

        if 200 != resp.status:
            raise InvalidResponse(
                "Error, server retunrned: {status}".format(status=resp.status)
            )

        try:
            data = await resp.json()
        except Exception as err:
            raise InvalidResponse(err)

        try:
            validate(data, ERR_JSONRPC20)
            return Response(**data)
        except ValidationError:
            # Passing data to validate response.
            # Good if does not valid to ERR_JSONRPC20 object.
            pass
        except Exception as err:
            raise InvalidResponse(err)

        try:
            validate(data, RSP_JSONRPC20)
            if id != data["id"]:
                raise InvalidResponse(
                    "Rsponse id {local} not equal {remote}".format(
                        local=id, remote=data["id"]
                    )
                )
        except Exception as err:
            raise InvalidResponse(err)

        if schem:
            try:
                validate(data["result"], schem)
            except ValidationError as err:
                raise InvalidResponse(err)
            except Exception as err:
                raise InternalError(err)

        return Response(**data)
